@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 wsproto/frame_protocol
 ~~~~~~~~~~~~~~~~~~~~~~
@@ -6,30 +5,45 @@ wsproto/frame_protocol
 WebSocket frame protocol implementation.
 """
 
-import itertools
 import os
 import struct
-from codecs import getincrementaldecoder
-from collections import namedtuple
+from codecs import getincrementaldecoder, IncrementalDecoder
 from enum import IntEnum
+from typing import Generator, List, NamedTuple, Optional, Tuple, TYPE_CHECKING, Union
 
-from .compat import unicode, Utf8Validator
+if TYPE_CHECKING:
+    from .extensions import Extension  # pragma: no cover
 
-try:
-    from wsaccel.xormask import XorMaskerSimple
-except ImportError:
 
-    class XorMaskerSimple:
-        def __init__(self, masking_key):
-            self._maskbytes = itertools.cycle(bytearray(masking_key))
+_XOR_TABLE = [bytes(a ^ b for a in range(256)) for b in range(256)]
 
-        def process(self, data):
-            maskbytes = self._maskbytes
-            return bytearray(b ^ next(maskbytes) for b in bytearray(data))
+
+class XorMaskerSimple:
+    def __init__(self, masking_key: bytes) -> None:
+        self._masking_key = masking_key
+
+    def process(self, data: bytes) -> bytes:
+        if data:
+            data_array = bytearray(data)
+            a, b, c, d = (_XOR_TABLE[n] for n in self._masking_key)
+            data_array[::4] = data_array[::4].translate(a)
+            data_array[1::4] = data_array[1::4].translate(b)
+            data_array[2::4] = data_array[2::4].translate(c)
+            data_array[3::4] = data_array[3::4].translate(d)
+
+            # Rotate the masking key so that the next usage continues
+            # with the next key element, rather than restarting.
+            key_rotation = len(data) % 4
+            self._masking_key = (
+                self._masking_key[key_rotation:] + self._masking_key[:key_rotation]
+            )
+
+            return bytes(data_array)
+        return data
 
 
 class XorMaskerNull:
-    def process(self, data):
+    def process(self, data: bytes) -> bytes:
         return data
 
 
@@ -39,8 +53,8 @@ class XorMaskerNull:
 PAYLOAD_LENGTH_TWO_BYTE = 126
 PAYLOAD_LENGTH_EIGHT_BYTE = 127
 MAX_PAYLOAD_NORMAL = 125
-MAX_PAYLOAD_TWO_BYTE = 2 ** 16 - 1
-MAX_PAYLOAD_EIGHT_BYTE = 2 ** 64 - 1
+MAX_PAYLOAD_TWO_BYTE = 2**16 - 1
+MAX_PAYLOAD_EIGHT_BYTE = 2**64 - 1
 MAX_FRAME_PAYLOAD = MAX_PAYLOAD_EIGHT_BYTE
 
 # MASK and PAYLOAD LEN are packed into a byte
@@ -60,7 +74,7 @@ class Opcode(IntEnum):
     RFC 6455, Section 5.2 - Base Framing Protocol
     """
 
-    #: Contiuation frame
+    #: Continuation frame
     CONTINUATION = 0x0
 
     #: Text message
@@ -78,7 +92,7 @@ class Opcode(IntEnum):
     #: Pong frame
     PONG = 0xA
 
-    def iscontrol(self):
+    def iscontrol(self) -> bool:
         return bool(self & 0x08)
 
 
@@ -192,21 +206,35 @@ NULL_MASK = struct.pack("!I", 0)
 
 
 class ParseFailed(Exception):
-    def __init__(self, msg, code=CloseReason.PROTOCOL_ERROR):
-        super(ParseFailed, self).__init__(msg)
+    def __init__(
+        self, msg: str, code: CloseReason = CloseReason.PROTOCOL_ERROR
+    ) -> None:
+        super().__init__(msg)
         self.code = code
 
 
-Header = namedtuple("Header", "fin rsv opcode payload_len masking_key".split())
+class RsvBits(NamedTuple):
+    rsv1: bool
+    rsv2: bool
+    rsv3: bool
 
 
-Frame = namedtuple("Frame", "opcode payload frame_finished message_finished".split())
+class Header(NamedTuple):
+    fin: bool
+    rsv: RsvBits
+    opcode: Opcode
+    payload_len: int
+    masking_key: Optional[bytes]
 
 
-RsvBits = namedtuple("RsvBits", "rsv1 rsv2 rsv3".split())
+class Frame(NamedTuple):
+    opcode: Opcode
+    payload: Union[bytes, str, Tuple[int, str]]
+    frame_finished: bool
+    message_finished: bool
 
 
-def _truncate_utf8(data, nbytes):
+def _truncate_utf8(data: bytes, nbytes: int) -> bytes:
     if len(data) <= nbytes:
         return data
 
@@ -222,17 +250,17 @@ def _truncate_utf8(data, nbytes):
     return data
 
 
-class Buffer(object):
-    def __init__(self, initial_bytes=None):
+class Buffer:
+    def __init__(self, initial_bytes: Optional[bytes] = None) -> None:
         self.buffer = bytearray()
         self.bytes_used = 0
         if initial_bytes:
             self.feed(initial_bytes)
 
-    def feed(self, new_bytes):
+    def feed(self, new_bytes: bytes) -> None:
         self.buffer += new_bytes
 
-    def consume_at_most(self, nbytes):
+    def consume_at_most(self, nbytes: int) -> bytes:
         if not nbytes:
             return bytearray()
 
@@ -240,31 +268,30 @@ class Buffer(object):
         self.bytes_used += len(data)
         return data
 
-    def consume_exactly(self, nbytes):
+    def consume_exactly(self, nbytes: int) -> Optional[bytes]:
         if len(self.buffer) - self.bytes_used < nbytes:
             return None
 
         return self.consume_at_most(nbytes)
 
-    def commit(self):
+    def commit(self) -> None:
         # In CPython 3.4+, del[:n] is amortized O(n), *not* quadratic
         del self.buffer[: self.bytes_used]
         self.bytes_used = 0
 
-    def rollback(self):
+    def rollback(self) -> None:
         self.bytes_used = 0
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.buffer)
 
 
-class MessageDecoder(object):
-    def __init__(self):
-        self.opcode = None
-        self.validator = None
-        self.decoder = None
+class MessageDecoder:
+    def __init__(self) -> None:
+        self.opcode: Optional[Opcode] = None
+        self.decoder: Optional[IncrementalDecoder] = None
 
-    def process_frame(self, frame):
+    def process_frame(self, frame: Frame) -> Frame:
         assert not frame.opcode.iscontrol()
 
         if self.opcode is None:
@@ -275,15 +302,18 @@ class MessageDecoder(object):
             raise ParseFailed("expected CONTINUATION, got %r" % frame.opcode)
 
         if frame.opcode is Opcode.TEXT:
-            self.validator = Utf8Validator()
             self.decoder = getincrementaldecoder("utf-8")()
 
         finished = frame.frame_finished and frame.message_finished
 
-        if self.decoder is not None:
-            data = self.decode_payload(frame.payload, finished)
-        else:
+        if self.decoder is None:
             data = frame.payload
+        else:
+            assert isinstance(frame.payload, (bytes, bytearray))
+            try:
+                data = self.decoder.decode(frame.payload, finished)
+            except UnicodeDecodeError as exc:
+                raise ParseFailed(str(exc), CloseReason.INVALID_FRAME_PAYLOAD_DATA)
 
         frame = Frame(self.opcode, data, frame.frame_finished, finished)
 
@@ -293,42 +323,33 @@ class MessageDecoder(object):
 
         return frame
 
-    def decode_payload(self, data, finished):
-        if self.validator is not None:
-            results = self.validator.validate(bytes(data))
-            if not results[0] or (finished and not results[1]):
-                raise ParseFailed(
-                    u"encountered invalid UTF-8 while processing"
-                    " text message at payload octet index %d" % results[3],
-                    CloseReason.INVALID_FRAME_PAYLOAD_DATA,
-                )
 
-        try:
-            return self.decoder.decode(data, finished)
-        except UnicodeDecodeError as exc:
-            raise ParseFailed(str(exc), CloseReason.INVALID_FRAME_PAYLOAD_DATA)
-
-
-class FrameDecoder(object):
-    def __init__(self, client, extensions=None):
+class FrameDecoder:
+    def __init__(
+        self, client: bool, extensions: Optional[List["Extension"]] = None
+    ) -> None:
         self.client = client
         self.extensions = extensions or []
 
         self.buffer = Buffer()
 
-        self.header = None
-        self.effective_opcode = None
-        self.masker = None
+        self.header: Optional[Header] = None
+        self.effective_opcode: Optional[Opcode] = None
+        self.masker: Union[None, XorMaskerNull, XorMaskerSimple] = None
         self.payload_required = 0
         self.payload_consumed = 0
 
-    def receive_bytes(self, data):
+    def receive_bytes(self, data: bytes) -> None:
         self.buffer.feed(data)
 
-    def process_buffer(self):
+    def process_buffer(self) -> Optional[Frame]:
         if not self.header:
             if not self.parse_header():
                 return None
+        # parse_header() sets these.
+        assert self.header is not None
+        assert self.masker is not None
+        assert self.effective_opcode is not None
 
         if len(self.buffer) < self.payload_required:
             return None
@@ -345,9 +366,10 @@ class FrameDecoder(object):
         payload = self.masker.process(payload)
 
         for extension in self.extensions:
-            payload = extension.frame_inbound_payload_data(self, payload)
-            if isinstance(payload, CloseReason):
-                raise ParseFailed("error in extension", payload)
+            payload_ = extension.frame_inbound_payload_data(self, payload)
+            if isinstance(payload_, CloseReason):
+                raise ParseFailed("error in extension", payload_)
+            payload = payload_
 
         if finished:
             final = bytearray()
@@ -370,7 +392,7 @@ class FrameDecoder(object):
 
         return frame
 
-    def parse_header(self):
+    def parse_header(self) -> bool:
         data = self.buffer.consume_exactly(2)
         if data is None:
             self.buffer.rollback()
@@ -386,14 +408,14 @@ class FrameDecoder(object):
         try:
             opcode = Opcode(opcode)
         except ValueError:
-            raise ParseFailed("Invalid opcode {:#x}".format(opcode))
+            raise ParseFailed(f"Invalid opcode {opcode:#x}")
 
         if opcode.iscontrol() and not fin:
             raise ParseFailed("Invalid attempt to fragment control frame")
 
         has_mask = bool(data[1] & MASK_MASK)
-        payload_len = data[1] & PAYLOAD_LEN_MASK
-        payload_len = self.parse_extended_payload_length(opcode, payload_len)
+        payload_len_short = data[1] & PAYLOAD_LEN_MASK
+        payload_len = self.parse_extended_payload_length(opcode, payload_len_short)
         if payload_len is None:
             self.buffer.rollback()
             return False
@@ -423,7 +445,9 @@ class FrameDecoder(object):
         self.payload_consumed = 0
         return True
 
-    def parse_extended_payload_length(self, opcode, payload_len):
+    def parse_extended_payload_length(
+        self, opcode: Opcode, payload_len: int
+    ) -> Optional[int]:
         if opcode.iscontrol() and payload_len > MAX_PAYLOAD_NORMAL:
             raise ParseFailed("Control frame with payload len > 125")
         if payload_len == PAYLOAD_LENGTH_TWO_BYTE:
@@ -451,7 +475,9 @@ class FrameDecoder(object):
 
         return payload_len
 
-    def extension_processing(self, opcode, rsv, payload_len):
+    def extension_processing(
+        self, opcode: Opcode, rsv: RsvBits, payload_len: int
+    ) -> None:
         rsv_used = [False, False, False]
         for extension in self.extensions:
             result = extension.frame_inbound_header(self, opcode, rsv, payload_len)
@@ -465,8 +491,8 @@ class FrameDecoder(object):
                 raise ParseFailed("Reserved bit set unexpectedly")
 
 
-class FrameProtocol(object):
-    def __init__(self, client, extensions):
+class FrameProtocol:
+    def __init__(self, client: bool, extensions: List["Extension"]) -> None:
         self.client = client
         self.extensions = [ext for ext in extensions if ext.enabled()]
 
@@ -475,10 +501,11 @@ class FrameProtocol(object):
         self._message_decoder = MessageDecoder()
         self._parse_more = self._parse_more_gen()
 
-        self._outbound_opcode = None
+        self._outbound_opcode: Optional[Opcode] = None
 
-    def _process_close(self, frame):
+    def _process_close(self, frame: Frame) -> Frame:
         data = frame.payload
+        assert isinstance(data, (bytes, bytearray))
 
         if not data:
             # "If this Close control frame contains no status code, _The
@@ -498,16 +525,6 @@ class FrameProtocol(object):
                 raise ParseFailed("remote CLOSE with local-only reason")
             if not isinstance(code, CloseReason) and code <= MAX_PROTOCOL_CLOSE_REASON:
                 raise ParseFailed("CLOSE with unknown reserved code")
-            validator = Utf8Validator()
-            if validator is not None:
-                results = validator.validate(bytes(data[2:]))
-                if not (results[0] and results[1]):
-                    raise ParseFailed(
-                        u"encountered invalid UTF-8 while"
-                        " processing close message at payload"
-                        " octet index %d" % results[3],
-                        CloseReason.INVALID_FRAME_PAYLOAD_DATA,
-                    )
             try:
                 reason = data[2:].decode("utf-8")
             except UnicodeDecodeError as exc:
@@ -519,7 +536,7 @@ class FrameProtocol(object):
 
         return Frame(frame.opcode, data, frame.frame_finished, frame.message_finished)
 
-    def _parse_more_gen(self):
+    def _parse_more_gen(self) -> Generator[Optional[Frame], None, None]:
         # Consume as much as we can from self._buffer, yielding events, and
         # then yield None when we need more data. Or raise ParseFailed.
 
@@ -540,17 +557,17 @@ class FrameProtocol(object):
 
             yield frame
 
-    def receive_bytes(self, data):
+    def receive_bytes(self, data: bytes) -> None:
         self._frame_decoder.receive_bytes(data)
 
-    def received_frames(self):
+    def received_frames(self) -> Generator[Frame, None, None]:
         for event in self._parse_more:
             if event is None:
                 break
             else:
                 yield event
 
-    def close(self, code=None, reason=None):
+    def close(self, code: Optional[int] = None, reason: Optional[str] = None) -> bytes:
         payload = bytearray()
         if code is None and reason is not None:
             raise TypeError("cannot specify a reason without a code")
@@ -565,16 +582,18 @@ class FrameProtocol(object):
 
         return self._serialize_frame(Opcode.CLOSE, payload)
 
-    def ping(self, payload=b""):
+    def ping(self, payload: bytes = b"") -> bytes:
         return self._serialize_frame(Opcode.PING, payload)
 
-    def pong(self, payload=b""):
+    def pong(self, payload: bytes = b"") -> bytes:
         return self._serialize_frame(Opcode.PONG, payload)
 
-    def send_data(self, payload=b"", fin=True):
+    def send_data(
+        self, payload: Union[bytes, bytearray, str] = b"", fin: bool = True
+    ) -> bytes:
         if isinstance(payload, (bytes, bytearray, memoryview)):
             opcode = Opcode.BINARY
-        elif isinstance(payload, unicode):
+        elif isinstance(payload, str):
             opcode = Opcode.TEXT
             payload = payload.encode("utf-8")
         else:
@@ -592,14 +611,16 @@ class FrameProtocol(object):
 
         return self._serialize_frame(opcode, payload, fin)
 
-    def _make_fin_rsv_opcode(self, fin, rsv, opcode):
-        fin = int(fin) << 7
-        rsv = (int(rsv.rsv1) << 6) + (int(rsv.rsv2) << 5) + (int(rsv.rsv3) << 4)
-        opcode = int(opcode)
+    def _make_fin_rsv_opcode(self, fin: bool, rsv: RsvBits, opcode: Opcode) -> int:
+        fin_bits = int(fin) << 7
+        rsv_bits = (int(rsv.rsv1) << 6) + (int(rsv.rsv2) << 5) + (int(rsv.rsv3) << 4)
+        opcode_bits = int(opcode)
 
-        return fin | rsv | opcode
+        return fin_bits | rsv_bits | opcode_bits
 
-    def _serialize_frame(self, opcode, payload=b"", fin=True):
+    def _serialize_frame(
+        self, opcode: Opcode, payload: bytes = b"", fin: bool = True
+    ) -> bytes:
         rsv = RsvBits(False, False, False)
         for extension in reversed(self.extensions):
             rsv, payload = extension.frame_outbound(self, opcode, rsv, payload, fin)
